# -------------------------------------------------------------------------- #
# 
# viterbi algorithm, find the best path
#
# -------------------------------------------------------------------------- #

"viterbi" <- 
function(x, p, a, du, D, pi, step){

    whichMax <- function(y){
        if(all(is.infinite(y))){
            return(NA)
        }else{
            return(which.max(y))
        }
    }

    check.input(x, p, a, du, D, pi)
    
    ## number of observiations
    n = length(x)
    
    ## number of states
    m = nrow(a)

    ## make sure p start from a value < step
    p = p - floor(min(p)/step)*step
    
    ## how many steps altogether
    Z = ceiling(max(p)/step)
    
    ## step indicator, every 'step' bp is one step
    stp = ceiling(p/step)
    
    ## every step must have at least 2 observations
    len = tapply(p, stp, length)
    if(any(len<2)){
        stop("At least one step has less than 2 observations\n")
    }
    
    ## indicator of end of every step
    dif = diff(stp)
    stp.end = c(which(dif>0), length(stp))

    ## log probability
    log.a  = log(a)
    log.du = log(du)
    log.pi = log(pi)

    ## results matrix
    prev = dura = xfix = matrix(NA, nrow=Z, ncol=m)
    logp = matrix(-1/0, nrow=Z, ncol=m)
    
    ## the previous step of state 1, 2 and 3 are fixed
    prev[,1] = 4
    prev[,2] = 3
    prev[,3] = 1

    ## initialization
    kp = 1:stp.end[1]
    ev = emiss(x[kp], p[kp], NULL, NULL, c(1,1,1,1))
    logp[1,] = log.pi + log.du[1,] + ev
    dura[1,is.finite(logp[1,])] = 1
    xfix[1,] = NA
    
    ## a vector used in the following steps
    v4 = numeric(2)
    
    ## recursion
    for(t in 2:Z){
        v = matrix(0, nrow=min(t,D), ncol=4)
        prev4 = numeric(min(t,D))
        for(d in 1:min(D, (t-1))){
            ## duration = 1:(t-1)
            kp = (stp.end[t-d]+1):stp.end[t]
            pfix = step*(t-d)
            em4 = emiss(x[kp], p[kp], xfix[t-d,4], pfix, c(1,0,0,0))
            em1 = emiss(x[kp], p[kp], xfix[t-d,1], pfix, c(0,0,1,0))
            em2 = emiss(x[kp], p[kp], xfix[t-d,2], pfix, c(0,0,0,1))
            em3 = emiss(x[kp], p[kp], xfix[t-d,3], pfix, c(0,1,0,1))

            v[d,1] = logp[t-d,4] + log.a[4,1] + log.du[d,1] + em4[1]
            v[d,2] = logp[t-d,3] + log.a[3,2] + log.du[d,2] + em3[2]
            v[d,3] = logp[t-d,1] + log.a[1,3] + log.du[d,3] + em1[3]
            
            v4[1]  = logp[t-d,2] + log.a[2,4] + em2[4]
            v4[2]  = logp[t-d,3] + log.a[3,4] + em3[4]
            v[d,4] = log.du[d,4] + max(v4)
            which4 = which.max(v4)
            
            ## it is possible that both v4[1] and v4[2] are -Inf
            if(length(which4)==0){
                prev4[d] = NA
            }else{
                prev4[d] = which.max(v4) + 1
            }
        }
        
        if(D>=t){
            ## duration = t
            kp = 1:stp.end[t]
            ev = emiss(x[kp], p[kp], NULL, NULL, c(1,1,1,1))
            v[t,] = log.pi + log.du[t,] + ev
        }
                
        dura[t,] = apply(v, 2, whichMax)
        
        for(i in 1:m){
            if(!is.na(dura[t,i])){
                d = dura[t,i]
                logp[t,i] = v[d,i]
                if(i==4){ prev[t,i] = prev4[d] }
                
                ## find the fix point
                if(d == t){stp.start = 1}
                else{ stp.start = stp.end[t-d]+1 }
                kp = stp.start:stp.end[t]
                xk = x[kp]
                pk = p[kp]
                if(d==t){
                    xfix1 = NULL
                    pfix1 = NULL
                }else{
                    xfix1 = xfix[t-d, prev[t,i]]
                    pfix1 = (t-d)*step
                }
                xfix[t,i] = get.end(xk, pk, t*step, i, xfix1, pfix1)
            }else{
                prev[t,i] = NA
            }
        }
        
    }
    
    path = rep(NA, Z)
    segment = matrix(NA, nrow=Z, ncol=2)
    seg = 0
    
    whichM = which.max(logp[Z,])
    if(length(whichM)==0){
        warning("no path found\n")
        return(NULL)
    }else{
        t2 = Z
        s2 = whichM
        seg = seg + 1
        segment[seg,1] = t2
        segment[seg,2] = xfix[t2,s2]
        d  = dura[t2, s2]
        s1 = prev[t2, s2]
        t1 = t2 - d
        path[(t1+1):t2]  = s2
    }
    
    while(t1>0){
        t2 = t1
        s2 = s1
        seg = seg + 1
        segment[seg,1] = t2
        segment[seg,2] = xfix[t2,s2]
        d  = dura[t2, s2]
        s1 = prev[t2, s2]
        if(is.na(d) || is.na(s2)){ break }
        t1 = t2 - d
        path[(t1+1):t2] = s2
    }
    segment = data.frame(segment[1:seg,])
    names(segment) = c("p", "x")
    bestPath = list(path=path, step=step, x=x, p=p, segment=segment)
    class(bestPath) = "hmm.path"
    
    bestPath
}
